"""
SitesDataFrame
-----------------

"""

import os
import json

import pandas
import folium
import numpy as np


class SitesDataFrame(pandas.DataFrame):
    """
    NEEDS TO HAVE THE FOLLOWING COLUMNS:

        site (CODE),
        name (NAME OF THE STATION),
        latitude,
        longitude
    """

    # pandas: disable=W0223

    def __init__(
            self,
            data=None,
            index=None,
            columns=None,
            dtype=None,
            copy=None,
            library: str = None,
            metadata: dict = None,
    ):

        super().__init__(
            data=data,
            index=index,
            columns=columns,
            dtype=dtype,
            copy=copy
        )

        self._validate(self)

        if metadata is None:
            metadata = {}
        object.__setattr__(self, "library", library)
        object.__setattr__(self, "metadata", metadata)

    @staticmethod
    def _validate(obj):
        """
        NEEDS TO HAVE THE FOLLOWING COLUMNS:

            site (CODE),
            name (NAME OF THE STATION),
            latitude,
            longitude
        """

        # verify there is a column latitude and a column longitude
        if "latitude" not in obj.columns or "longitude" not in obj.columns:
            raise AttributeError("Must have 'latitude' and 'longitude'.")
        # verify there is a column site and a column name
        if "site" not in obj.columns or "name" not in obj.columns:
            raise AttributeError("Each site must be identify by a code 'site'"
                                 + " and its name 'name'.")

    @staticmethod
    def open_from(
            data_fl: str = None,
            metadata_fl: str = None,
            folder_name: str = None
    ):
        """
        """

        if data_fl is None:
            if folder_name is None:
                raise KeyError("Not correct file path")
            data_fl = folder_name + "data.pkl"
        if metadata_fl is None:
            if folder_name is None:
                raise KeyError("Not correct file path")
            metadata_fl = folder_name + "metadata.json"

        return SitesDataFrame(
            data=pandas.read_csv(data_fl),
            library="pyaemet",
            metadata=json.load(metadata_fl)
            )

    def save(self, folder_name: str, extension: str = 'pickle'):
        """
        """

        if not os.path.exists(folder_name):
            os.makedirs(folder_name)

        if extension == 'pickle':
            self.to_pickle(folder_name+"data.pkl")
        elif extension == 'csv':
            self.to_csv(folder_name+"data.csv")

        with open(folder_name+"metadata.json", 'w') as file:
            json.dump(self.metadata, file, indent=4)

    def copy(self, deep=True):
        """ Copy object """

        return SitesDataFrame(data=super().copy(deep),
                              library=self.library,
                              metadata=self.metadata)

    def as_dataframe(self):
        return super().copy(True)

    @property
    def map(self):
        """
        plot map with the sites location
        """

        if self.empty:
            return folium.Map()

        index_lat, = np.where(self.columns == "latitude")[0]
        index_lon, = np.where(self.columns == "longitude")[0]

        latitudes = self._get_column_array(index_lat)
        longitudes = self._get_column_array(index_lon)

        center = [np.mean([np.min(latitudes), np.max(latitudes)]),
                  np.mean([np.min(longitudes), np.max(longitudes)])]

        mapa = folium.Map(location=center, zoom_start=3)

        for i in self.index:
            popup = ("<strong>Site:</strong> %s<br>"
                     % (self._get_value(i, "site")) +
                     "<strong>Name:</strong> %s"
                     % (self._get_value(i, "name")))
            folium.Marker([self._get_value(i, "latitude"),
                           self._get_value(i, "longitude")],
                          popup=folium.Popup(popup, max_width=480),
                          tooltip="Click me!").add_to(mapa)

        return mapa

    def filter_in(self, **kwargs):
        """
        """

        kwargs.update(
            {k: [v] for k, v in kwargs.items() if not isinstance(v, list)})

        if any(not isinstance(x, list) for x in kwargs.values()):
            raise(TypeError("only list-like objects are allowed to be passed" +
                            " to filter_in(), you passed a [str]"))
        if any(x not in self.columns for x in kwargs):
            raise(KeyError("The keys passed to filter_in() does not match " +
                           "with SitesDataFrame columns"))

        sites = super().copy()
        for ky, vl in kwargs.items():
            sites = sites.__getitem__(sites.__getitem__(ky).isin(vl))

        return SitesDataFrame(data=sites,
                              library=self.library,
                              metadata=self.metadata)

    def filter_at(
            self,
            latitude: float,
            longitude: float,
            n_near: int = 100,
            max_distance: float = 6237.0,
    ):
        """
        """

        sites_distance = self.calc_distance(latitude, longitude)

        sites_distance = sites_distance.__getitem__(
            sites_distance.__getitem__("distance") <=
            max_distance).sort_values(by=['distance'],
                                      ascending=True)[:(n_near)]

        return NearSitesDataFrame(ref_point=[latitude, longitude],
                                  data=sites_distance, metadata=self.metadata)

    def calc_distance(
            self,
            latitude: float,
            longitude: float,
            radius: float = 6371.0,
    ):

        """
        Calculate distance between locations given in latitudes and
        longitudes coordinates. The distance is calculated using the
        following equation:

            dist = radius *
                arcocos{cos(lat1 - lat2) -
                        cos(lat1)*cos(lat2)*[1 cos(long1 - long2)]}

        :param coor1: list of coordinates of first location
            e.g.: [latitude,longitude]
        :param corr2: list of coordinates of second location
            e.g.: [latitude,longitude]
        :param radius: earth radius

        :returns: distance between the two locations in kilometers
            (Due to radius is given in km)
        """

        coor1 = pandas.DataFrame(np.deg2rad(self.__getitem__(["latitude",
                                                              "longitude"])))
        latitude, longitude = np.deg2rad(latitude), np.deg2rad(longitude)

        new_data = self.copy()

        new_data.__setitem__(key="distance", value=coor1.apply(
            lambda row: radius *
            np.arccos(np.cos(row["latitude"] - latitude) -
                      np.cos(row["latitude"]) * np.cos(latitude) *
                      (1 - np.cos(row["longitude"] - longitude))
                      ), axis=1))

        return new_data


class NearSitesDataFrame(SitesDataFrame):
    """
    MUST HAVE THE FOLLOWING COLUMNS:

        site (CODE),
        name (NAME OF THE STATION),
        latitude,
        longitude,
        distance (TO THE REFERENCE POINT)
    """

    def __init__(
            self,
            ref_point: [float, float],
            data=None,
            index=None,
            columns=None,
            dtype=None,
            copy=None,
            library=None,
            metadata=None,
    ):
        """
        """

        super().__init__(
            data=data,
            index=index,
            columns=columns,
            dtype=dtype,
            copy=copy,
            library=library,
            metadata=metadata,
        )

        self.metadata.update({"Reference Point": {"latitude": ref_point[0],
                                                  "longitude": ref_point[1]
                                                  }
                              })

    @staticmethod
    def _validate(obj):
        """
        MUST HAVE THE FOLLOWING COLUMNS:

            site (CODE),
            name (NAME OF THE STATION),
            latitude,
            longitude
        """

        # verify there is a column latitude and a column longitude
        if "latitude" not in obj.columns or "longitude" not in obj.columns:
            raise AttributeError("Must have 'latitude' and 'longitude'.")
        # verify there is a column site and a column name
        if "site" not in obj.columns or "name" not in obj.columns:
            raise AttributeError("Each site must be identify by a code 'site'"
                                 + " and its name 'name'.")
        if "distance" not in obj.columns:
            raise AttributeError("NearSitesDataFrame must include the"
                                 + " 'distance' column")

    @property
    def map(self):
        """
        plot map with the sites location
        """

        index_lat, = np.where(self.columns == "latitude")[0]
        index_lon, = np.where(self.columns == "longitude")[0]

        latitudes = self._get_column_array(index_lat)
        longitudes = self._get_column_array(index_lon)

        center = [np.mean([np.min(latitudes), np.max(latitudes)]),
                  np.mean([np.min(longitudes), np.max(longitudes)])]

        mapa = folium.Map(location=center, zoom_start=3)
        folium.LayerControl().add_to(mapa)

        for i in self.index:
            popup = ("<strong>Site:</strong> %s<br>"
                     % (self._get_value(i, "site")) +
                     "<strong>Name:</strong> %s<br>"
                     % (self._get_value(i, "name")) +
                     "<strong>Distance:</strong> %.2f"
                     % (self._get_value(i, "distance")))

            folium.Marker([self._get_value(i, "latitude"),
                           self._get_value(i, "longitude")],
                          popup=folium.Popup(popup, max_width=480),
                          tooltip="Click me!").add_to(mapa)

        folium.Marker([self.metadata["Reference Point"]["latitude"],
                       self.metadata["Reference Point"]["longitude"]],
                      icon=folium.Icon(color="red"),
                      popup=folium.Popup("Reference Point", max_width=480),
                      tooltip="Click me!").add_to(mapa)

        return mapa

    def as_dataframe(self):
        return super().copy(True)

    def sort_values(self, **kwargs):
        return NearSitesDataFrame(
            data=super().sort_values(**kwargs),
            ref_point=[self.metadata["Reference Point"]["latitude"],
                       self.metadata["Reference Point"]["longitude"],
                       ],
            metadata=self.metadata
            )



""" AEMET API MODULE
-----------------

Python module to operate with AEMET OpenData API REST

:author Jaimedgp
"""

from datetime import date, datetime

import requests
import pandas as pd

from .types_classes.sites import SitesDataFrame

from .utilities.coordinates import transform_coordinates, get_site_address
from .utilities.dictionaries import SITES_TRANSLATION, OBSERVATIONS_TRANSLATION
from .utilities.curation import (
    update_fields,
    decimal_notation,
    convert_hours,
    remove_newline
    )



class _AemetApiRequest():
    """ Class to download data using AEMET api"""

    def __init__(self, apikey):
        """ Get the needed API key"""

        self.main_url = "https://opendata.aemet.es/opendata/api/"
        self._params = {"api_key": apikey}
        self._headers = {"cache-control": "no-cache",
                         "Accept": "application/json",
                         "Content-Type": "application/json",
                         }

    def _aemet_request(self, url):
        """
        """

        response = requests.request("GET",
                                    self.main_url+url,
                                    headers=self._headers,
                                    params=self._params
                                    )

        if response.ok:
            """
            Possible errores:
            -----------------

            429:
            404:
            401:
            """
            if response.text == "":
                return [{},
                        {"status":
                            "Nothing returned. Please check the API key"}
                       ]
            elif response.json()["estado"] == 200:
                data_url = response.json()["datos"]
                metadata_url = response.json()["metadatos"]

                return [requests.request("GET",
                                         data_url, headers=self._headers) \
                                .json(),
                        requests.request("GET",
                                         metadata_url,
                                         headers=self._headers) \
                                .json()
                        ]

        return [ {}, response.json() ]


class ClimaValues(_AemetApiRequest):
    """ Class to download climatological data using AEMET api"""

    def __init__(self, apikey):
        """ Get the needed API key"""

        super().__init__(apikey)
        self.main_url += "valores/climatologicos/"

    def get_sites_info(self, old_dataframe: SitesDataFrame):
        """
        """

        data, metadata = self._aemet_request(url=("inventarioestaciones/" +
                                                  "todasestaciones/"))

        if not bool(data):
            return pd.DataFrame(columns=SITES_TRANSLATION.keys()), metadata

        data = pd.DataFrame(data) \
                 .rename(columns={v["id"]: k
                                  for k, v in SITES_TRANSLATION.items()}) \
                 .apply(transform_coordinates) \
                 .astype({k: v["dtype"]
                          for k, v in SITES_TRANSLATION.items()
                          if k in data})

        if (not all(data.columns.isin(old_dataframe.columns)) or
                (not data.equals(old_dataframe.loc[:, data.columns]))):

            data = data.merge(data.apply(get_site_address, axis=1,
                                         result_type="expand")
                                  .drop_duplicates(),
                              on=["latitude", "longitude"],
                              how='left'
                              ).drop_duplicates()

        else:
            old_dataframe.metadata.update({"access_date":
                                           datetime.now().isoformat()})
            return old_dataframe

        metadata = {k+"_aemet": v for k, v in metadata.items()}
        metadata["access_date"] = datetime.now().isoformat()
        metadata["fields"] = update_fields(data.columns,
                                           metadata.pop("campos_aemet"),
                                           SITES_TRANSLATION)

        return remove_newline(data), metadata

    def get_observations(
            self,
            fechaIniStr: date,
            fechaFinStr: date,
            idema: str
    ):
        """ Docstring """

        params = {"fechaIniStr": fechaIniStr.strftime("%Y-%m-%dT%H:%M:%SUTC"),
                  "fechaFinStr": fechaFinStr.strftime("%Y-%m-%dT%H:%M:%SUTC"),
                  "idema": idema
                  }

        data, metadata = self._aemet_request(url=("diarios/datos/fechaini/" +
                                                  "{fechaIniStr}/fechafin/" +
                                                  "{fechaFinStr}/estacion/" +
                                                  "{idema}"
                                                  ).format(**params))

        if not bool(data):
            return (pd.DataFrame(columns=OBSERVATIONS_TRANSLATION.keys()),
                    metadata)

        data = pd.DataFrame(data) \
                 .drop(["nombre", "provincia"], axis=1) \
                 .rename(columns={v["id"]: k
                                  for k, v in OBSERVATIONS_TRANSLATION.items()
                                  }) \
                 .replace({"Ip": "0,05", "Varias": "-1", "Acum": None}) \
                 .apply(decimal_notation, axis=1)

        data = data.astype({k: v["dtype"]
                            for k, v in OBSERVATIONS_TRANSLATION.items()
                            if k in data.columns}) \
                   .apply(convert_hours)

        metadata = {k+"_aemet": v for k, v in metadata.items()}
        metadata["access_date"] = datetime.now().isoformat()
        metadata["fields"] = update_fields(data,
                                           metadata.pop("campos_aemet"),
                                           OBSERVATIONS_TRANSLATION)

        return remove_newline(data), metadata



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['pyaemet', 'pyaemet.static', 'pyaemet.types_classes', 'pyaemet.utilities']

package_data = \
{'': ['*'], 'pyaemet.static': ['sites/*']}

install_requires = \
['folium>=0.14.0,<0.15.0',
 'geocoder>=1.38.1,<2.0.0',
 'matplotlib>=3.7.1,<4.0.0',
 'numpy>=1.24.3,<2.0.0',
 'pandas>=2.0.1,<3.0.0',
 'requests>=2.28.2,<3.0.0',
 'tqdm>=4.65.0,<5.0.0']

setup_kwargs = {
    'name': 'pyaemet',
    'version': '1.1.0rc0',
    'description': 'Python module to interact with the AEMET API to download meteorological data',
    'long_description': '# pyAEMET\n\n\n[![PyPI Latest Release](https://img.shields.io/pypi/v/pyaemet.svg)](https://pypi.org/project/pyaemet/)\n[![DOI](https://zenodo.org/badge/DOI/10.5281/zenodo.5655307.svg)](https://doi.org/10.5281/zenodo.5655307)\n[![License](https://img.shields.io/pypi/l/pandas.svg)](https://github.com/jaimedgp/pyAEMET/blob/main/LICENSE)\n[![Downloads](https://static.pepy.tech/personalized-badge/pyaemet?period=month&units=international_system&left_color=gray&right_color=orange&left_text=PyPI%20downloads%20per%20month)](https://pepy.tech/project/pyaemet)\n\nA python library developed to download daily climatological values from the Spanish National\nMeteorological Agency (AEMET) through its OpenData API. The library contains several methods\nto facilitate downloading and filtering the climatological data.\n\n> The information that this library collects and uses is property of the Spanish State\n> Meteorological Agency, available through its AEMET OpenData REST API.\n\n\n## Installation\n``` bash\n$ pip install pyaemet\n```\nTo use the pyAEMET module, you need to get an API key from the AEMET (Spanish State Meteorological\nAgency) OpenData platform. You can apply for a key [here](https://opendata.aemet.es/centrodedescargas/altaUsuario).\n\n## Usage\n\nOnce the module is installed and you have your API key, you can start using the module by\nimporting it in your Python script. To use the module\'s functions, you need to initialize\nthe client with your API key:\n\n```python\nimport pyaemet\n\naemet = pyaemet.AemetClima(api_key)\n```\n\nThe `AemetClima` class takes an API key as a parameter in its constructor and allows you to get\ninformation about the available monitoring sites, filter sites based on different parameters\n(e.g., city, province, autonomous community), and get nearby sites to a specific location.\n\nHere is a summary of some of the methods provided by the `AemetClima` class:\n\n* **`sites_info`**: Retrieves information about all the available monitoring sites. The method\nreturns an instance of the `SitesDataFrame` class, which is a subclass of the pandas `DataFrame`.\n```python\naemet.sites_info(update=True)\n```\n\n* **`sites_in`**: Filters the available monitoring sites based on specified parameters\n(e.g., city, province, autonomous community). The method returns an instance of the `SitesDataFrame` class.\n```python\naemet.sites_in(subregion="Cantabria")\n```\n![image](https://github.com/Jaimedgp/pyAEMET/raw/main/docs/screenshots/sites_cantabria.png)\n\n* **`near_sites`**: Retrieves the ``n_near`` monitoring sites closest to a specified latitude and longitude,\nwithin a maximum distance of `max_distance` kilometers. The method returns an instance of the\n`NearSitesDataFrame` class.\n```python\naemet.near_sites(latitude=43.47,\n                 longitude=-3.798,\n                 n_near=5, max_distance=50)\n```\n![image](https://github.com/Jaimedgp/pyAEMET/raw/main/docs/screenshots/near_sites.png)\n\n* **`sites_curation`**: Retrieves the amount of available data of certain `variables` in the monitoring `sites` in a period of time defined by\n    `start_dt` and `end_dt`. The function returns a `SitesDataFrame` or `NearSitesDataFrame` (depends of the type of the `sites` parameter given)\n    with a column with the average `amount` between all `variables` and `has_enough` boolean if the amount is greater or equal to a `threshold`.\n\n* **`daily_clima`**: Retrieves daily climate data for a given ``site`` or a list of sites over a\nspecified date range defined by `start_dt` and `end_dt`. The function returns a\n`ObservationsDataFrame` object, which is a data structure that holds the retrieved climate data\nalong with any associated metadata.\n```python\nimport datetime\naemet.daily_clima(site=aemet.sites_in(city="Santander"),\n                  start_dt=datetime.date(2022, 6, 3),\n                  end_dt=datetime.date.today())\n```\n\nThe module also provides three deprecated methods `estaciones_info`, `estaciones_loc` and `clima_diaria`\nthat perform similar functionality as the `sites_info`, `sites_in` and `daily_clima` methods, respectively.\n\nYou can find the complete documentation of the module\'s functions in the GitHub repository,\nunder the docs directory.\n\n## FAQ\n## Contributing\n## References\n* ["Estimating changes in air pollutant levels due to COVID-19 lockdown measures based on a business-as-usual prediction scenario using data mining models: A case-study for urban traffic sites in Spain"](https://doi.org/10.1016/j.scitotenv.2022.153786), submitted to Environmental Software & Modelling by [J. González-Pardo](https://orcid.org/0000-0001-7268-9933) et al. (2021)\n',
    'author': 'Jaimedgp',
    'author_email': 'jaime.diez.gp@gmail.com',
    'maintainer': 'CarmenGBM',
    'maintainer_email': 'carmen.garcia.be96@gmail.com',
    'url': 'https://github.com/jaimedgp/pyAEMET',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
